#include "RdmaClientEndpoint.hpp"

int RdmaClientEndpoint::CONN_STATE_INITIALIZED = 0;
int RdmaClientEndpoint::CONN_STATE_ADDR_RESOLVED = 1;
int RdmaClientEndpoint::CONN_STATE_ROUTE_RESOLVED = 2;
int RdmaClientEndpoint::CONN_STATE_RESOURCES_ALLOCATED = 3;
int RdmaClientEndpoint::CONN_STATE_CONNECTED = 4;
int RdmaClientEndpoint::CONN_STATE_PARTIAL_CLOSED = 5;
int RdmaClientEndpoint::CONN_STATE_CLOSED = 6;

RdmaClientEndpoint::RdmaClientEndpoint(struct rdma_cm_id *id, RdmaEndpointGroup *group, int sendQueueSize, int recvQueueSize,
                                       int sendMsgSize, int recvMsgSize, int maxInLine, int timeout)
    : _cm_id(id), _group(group), _sendQueueSize(sendQueueSize), _recvQueueSize(recvQueueSize),
      _sendMsgSize(sendMsgSize), _recvMsgSize(recvMsgSize), _maxInLine(maxInLine), _timeoutMs(timeout)
{
    _state = CONN_STATE_INITIALIZED;
}

void RdmaClientEndpoint::connect(const char *ip, const char *port, const char *connData)
{
    _connData = connData;
    if (_state != CONN_STATE_INITIALIZED)
    {
        CPPLog::LOG_DEBUG("RdmaClientEndpoint : connect state not initialized");
        return;
    }
    int ret;
    CPPLog::LOG_DEBUG("RdmaClientEndpoint :  step2 getaddrinfo");
    struct addrinfo *addr;
    ret = getaddrinfo(ip, port, NULL, &addr);
    if (ret)
    {
        CPPLog::LOG_ERROR("RdmaServerEndpointGroup : get_addr_info failed");
    }

    CPPLog::LOG_DEBUG("RdmaClientEndpoint :  step2 resolve addr");
    ret = rdma_resolve_addr(_cm_id, NULL, addr->ai_addr, _timeoutMs);
    if (ret)
    {
        CPPLog::LOG_ERROR("unable to resolve addr");
        return;
    }
    CPPLog::LOG_DEBUG("RdmaClientEndpoint :  step2 resolve addr resolved");
    _state = CONN_STATE_ADDR_RESOLVED;
}

void RdmaClientEndpoint::connect()
{
    if (_connData != NULL)
    {
        struct rdma_conn_param conn_param;
        memset(&conn_param, 0, sizeof(conn_param));
        conn_param.responder_resources = 1;
        conn_param.initiator_depth = 1;
        conn_param.retry_count = 7;
        conn_param.rnr_retry_count = 7;
        conn_param.private_data = _connData;
        conn_param.private_data_len = strlen(_connData);
        rdma_connect(_cm_id, &conn_param);
    }
    else
    {
        rdma_connect(_cm_id, NULL);
    }
}


bool RdmaClientEndpoint::isConnected()
{
    return _state == CONN_STATE_CONNECTED;
}


void RdmaClientEndpoint::processCmEvent(struct rdma_cm_event *event)
{
    if (event->event == RDMA_CM_EVENT_ADDR_RESOLVED && event->id != NULL)
    {
        CPPLog::LOG_DEBUG("RdmaClientEndpoint : step3 resolve_route");
        createResources();
        rdma_resolve_route(_cm_id, _timeoutMs);
    }
    else if (event->event == RDMA_CM_EVENT_ROUTE_RESOLVED && event->id != NULL)
    {
        registerMemory();
        CPPLog::LOG_DEBUG("RdmaClientEndpoint : step5 connect Request");
        connect();
    }
    else if (event->id != NULL && event->event == RDMA_CM_EVENT_ESTABLISHED)
    {
        CPPLog::LOG_DEBUG("RdmaClientEndpoint : step6 Connected");
        _state = CONN_STATE_CONNECTED;
    }
    else if (event->id != NULL && event->event == RDMA_CM_EVENT_DISCONNECTED)
    {
        CPPLog::LOG_DEBUG("RdmaClientEndpoint : step7 Closed");
        completeClose();
    }
    else
    {
        std::ostringstream ss;
        ss << "RdmaClientEndpoint : Not able to procces CM Event ";
        ss << rdma_event_str(event->event) << " " << event->id << " " << event->listen_id;
        CPPLog::LOG_ERROR(ss);
    }
}

void RdmaClientEndpoint::close()
{
    if (_state != CONN_STATE_CONNECTED)
    {
        CPPLog::LOG_ERROR("RdmaClientEndpoint : close invalid state");
        return;
    }
    _state = CONN_STATE_PARTIAL_CLOSED;
    int ret = rdma_disconnect(_cm_id);
    if (ret)
    {
        CPPLog::LOG_ERROR("RdmaClientEndpoint : rdma_disconnect failed");
    }
}

void RdmaClientEndpoint::completeClose()
{
    if (_state != CONN_STATE_PARTIAL_CLOSED)
    {
        CPPLog::LOG_ERROR("RdmaClientEndpoint : completeClose invalid state");
        //return;
    }
    _state = CONN_STATE_CLOSED;
    // delete _sendBuffers;
    rdma_dereg_mr(_sendMr);
    rdma_dereg_mr(_recvMr);
    delete[] _sendBuff;
    delete[] _recvBuff;
    rdma_destroy_qp(_cm_id);
    rdma_destroy_id(_cm_id);
    CPPLog::LOG_INFO("RdmaClientEndpoint : close connection");
}

void RdmaClientEndpoint::registerMemory()
{
    if (_state != CONN_STATE_ROUTE_RESOLVED)
    {
        CPPLog::LOG_ERROR("RdmaClientEndpoint : createResource address not resolved");
        return;
    }
    _sendBuff = new char[(_sendMsgSize+MessageHeaderSize) * _sendQueueSize];
    if (_sendBuff == NULL)
    {
        CPPLog::LOG_ERROR("RdmaClientEndpoint :  sendBuff allocation failed");
        return;
    }

    _sendMr = rdma_reg_msgs(_cm_id, reinterpret_cast<void *>(_sendBuff),
                            _sendMsgSize * _sendQueueSize);
    if (_sendMr == NULL)
    {
        CPPLog::LOG_ERROR("RdmaClientEndpoint :  sendMr reg failed");
        return;
    }

    _recvBuff = new char[(MessageHeaderSize + _recvMsgSize) * _recvQueueSize];
    if (_recvBuff == NULL)
    {
        CPPLog::LOG_ERROR("RdmaClientEndpoint :  recvBuff allocation failed\n");
        return;
    }

    _recvMr = rdma_reg_msgs(_cm_id, reinterpret_cast<void *>(_recvBuff),
                            _recvMsgSize * _recvQueueSize);

    if (_recvMr == NULL)
    {
        CPPLog::LOG_ERROR("RdmaClientEndpoint :  recvMr reg failed\n");
        return;
    }

    for (int i = 0; i < _recvQueueSize; i++)
    {
        char *buffer = _recvBuff + i * _recvMsgSize;
        rdma_post_recv(_cm_id, reinterpret_cast<void *>(buffer), reinterpret_cast<void *>(buffer),
                       _recvMsgSize, _recvMr);
    }
    for (int i = 0; i < _sendQueueSize; i++)
    {
        char *buffer = _sendBuff + i * _recvMsgSize;
        std::unique_lock<std::mutex> lock(_sendBuffersM);
        _sendBuffers.push(buffer);
        lock.unlock();
    }

    _state = CONN_STATE_RESOURCES_ALLOCATED;
}
void RdmaClientEndpoint::createResources()
{
    if (_state != CONN_STATE_ADDR_RESOLVED)
    {
        CPPLog::LOG_ERROR("RdmaClientEndpoint : createResource address not resolved");
        return;
    }
    _protectionDomain = ibv_alloc_pd(_cm_id->verbs);
    if (_protectionDomain == NULL)
    {
        CPPLog::LOG_ERROR("RdmaClientEndpoint : ibv_alloc_pd failed ");
        return;
    }
    struct ibv_cq *completionQueue = _group->createCq(_cm_id);
    struct ibv_qp_init_attr qp_init_attr;
    memset(&qp_init_attr, 0, sizeof(qp_init_attr));
    // This is used to set endpoint address with qp
    qp_init_attr.qp_context = (void *)this;
    // if not set 0, all work requests submitted to SQ will always generate a Work Completion
    qp_init_attr.sq_sig_all = 1;
    // completion queue can be shared or you can use distinct completion queues.
    qp_init_attr.send_cq = completionQueue;
    qp_init_attr.recv_cq = completionQueue;
    qp_init_attr.qp_type = IBV_QPT_RC;
    // increase if you want to keep more send work requests in the SQ.
    qp_init_attr.cap.max_send_wr = _sendQueueSize;
    // increase if you want to keep more receive work requests in the RQ.
    qp_init_attr.cap.max_recv_wr = _recvQueueSize;
    // increase if you allow send work requests to have multiple scatter gather entry (SGE).
    qp_init_attr.cap.max_send_sge = 1;
    // increase if you allow receive work requests to have multiple scatter gather entry (SGE).
    qp_init_attr.cap.max_recv_sge = 1;
    qp_init_attr.cap.max_inline_data = _maxInLine;

    //_queuePair = ibv_create_qp(_protectionDomain, &qp_init_attr);
    int ret = rdma_create_qp(_cm_id, _protectionDomain, &qp_init_attr);
    if (ret)
    {
        CPPLog::LOG_ERROR("RdmaClientEndpoint : ibv_create_cq failed");
    }
    if (_cm_id->pd == NULL)
    {
        CPPLog::LOG_ERROR("RdmaClientEndpoint : pd not set");
        _cm_id->pd = _protectionDomain;
    }

    _state = CONN_STATE_ROUTE_RESOLVED;
}

int RdmaClientEndpoint::sendMessage(const char *buffer, int size)
{
    if (size > _sendMsgSize)
    {
        std::ostringstream ss;
        ss<<"Large Message size "<<size;
        ss<<" send buffer size " << _sendMsgSize;
        CPPLog::LOG_ERROR(ss);
        return -1;
    }
        
    std::unique_lock<std::mutex> lock(_sendBuffersM);
    if (_sendBuffers.size() == 0)
        return -1;
    char *sendBuffer = _sendBuffers.front();
    _sendBuffers.pop();
    lock.unlock();

    memcpy(sendBuffer, buffer, size);
    return rdma_post_send(_cm_id, sendBuffer, sendBuffer, size, _sendMr, 0);
}

void RdmaClientEndpoint::processSendComp(struct ibv_wc wc)
{
    std::unique_lock<std::mutex> lock(_sendBuffersM);
    _sendBuffers.push((char *)wc.wr_id);
}
void RdmaClientEndpoint::processRecvComp(struct ibv_wc wc)
{
    char *data = new char[wc.byte_len];
    memcpy(data, (void *)wc.wr_id, wc.byte_len);
    rdma_post_recv(_cm_id, (void *)wc.wr_id, (void *)wc.wr_id, _recvMsgSize, _recvMr);

    struct MessageHeader *response = (struct MessageHeader *)data;
    /*std::ostringstream ss;
    ss << "Recieve response for id " << response->id << " status " << response->status;
    ss << " size " << response->valueSize << std::endl;
    CPPLog::LOG_INFO(ss);*/
    if (response->type != MessageType::INVALIDATE)
    {
        auto it = futures.find(response->id);
        if (it == futures.end())
        {
            std::ostringstream ss;
            ss << "Recv completion for invalid id" << response->id;
            CPPLog::LOG_DEBUG(ss);
        }
        it->second->put(data);
    }
}

RdmaFuture *RdmaClientEndpoint::put(const char *key, int keySize, const char *value, int valueSize)
{   
    if (keySize + valueSize + (int)MessageHeaderSize > _sendMsgSize)
        return nullptr;
    std::unique_lock<std::mutex> lock(_sendBuffersM);
    if (_sendBuffers.size() == 0)
        return nullptr;
    char *sendBuffer = _sendBuffers.front();
    _sendBuffers.pop();
    lock.unlock();
    struct MessageHeader *request = (struct MessageHeader *)sendBuffer;
    request->id = _requestId.fetch_add(1, std::memory_order_relaxed);
    request->type = MessageType::PUT;
    request->keySize = keySize;
    request->valueSize = valueSize;
    memcpy(sendBuffer + MessageHeaderSize, key, keySize);
    memcpy(sendBuffer + MessageHeaderSize + keySize, value, valueSize);
    rdma_post_send(_cm_id, sendBuffer, sendBuffer, MessageHeaderSize + keySize + valueSize,
                   _sendMr, 0);
    RdmaFuture *future = new RdmaFuture(request->id);
    futures[request->id] = future;
    std::cout<<"created future\n";
    return future;
}
RdmaFuture *RdmaClientEndpoint::get(const char *key, int keySize)
{
    if (keySize + (int)MessageHeaderSize > _sendMsgSize)
        return nullptr;
    std::unique_lock<std::mutex> lock(_sendBuffersM);
    if (_sendBuffers.size() == 0)
        return nullptr;
    char *sendBuffer = _sendBuffers.front();
    _sendBuffers.pop();
    lock.unlock();
    struct MessageHeader *request = (struct MessageHeader *)sendBuffer;
    request->id = _requestId.fetch_add(1, std::memory_order_relaxed);
    request->type = MessageType::GET;
    request->keySize = keySize;
    memcpy(sendBuffer + MessageHeaderSize, key, keySize);
    rdma_post_send(_cm_id, sendBuffer, sendBuffer, MessageHeaderSize + keySize,
                   _sendMr, 0);
    RdmaFuture *future = new RdmaFuture(request->id);
    futures[request->id] = future;
    return future;
}
RdmaFuture *RdmaClientEndpoint::deleteKey(const char *key, int keySize)
{
    if (keySize + (int)MessageHeaderSize > _sendMsgSize)
        return nullptr;
    std::unique_lock<std::mutex> lock(_sendBuffersM);
    if (_sendBuffers.size() == 0)
        return nullptr;
    char *sendBuffer = _sendBuffers.front();
    _sendBuffers.pop();
    lock.unlock();
    struct MessageHeader *request = (struct MessageHeader *)sendBuffer;
    request->id = _requestId.fetch_add(1, std::memory_order_relaxed);
    request->type = MessageType::DELETE;
    request->keySize = keySize;
    memcpy(sendBuffer + MessageHeaderSize, key, keySize);
    rdma_post_send(_cm_id, sendBuffer, sendBuffer, MessageHeaderSize + keySize,
                   _sendMr, 0);
    RdmaFuture *future = new RdmaFuture(request->id);
    futures[request->id] = future;
    return future;
}
